import contextlib
import logging
import os

# converts string in form [cos|s3]://access_key_id:secret_access_key@endpoint/bucket/path to
# access_key_id, secret_access_key, endpoint, path - path includes bucket name
def explode_connection_string(cs):
    if cs is None:
        return None
    if cs.startswith('cos') or cs.startswith('s3'):
        buffer=cs.split('://')[1]
        access_key_id=buffer.split('@')[0].split(':')[0]
        secret_access_key=buffer.split('@')[0].split(':')[1]
        endpoint=f"https://{buffer.split('@')[1].split('/')[0]}"
        path='/'.join(buffer.split('@')[1].split('/')[1:])
        return (access_key_id, secret_access_key, endpoint, path)
    else:
        return (None, None, None, cs)
        # TODO consider cs as secret and grab connection string from kubernetes
    

def run_and_log(cos_conn, log_folder, task_id, command_array):
    log_root_name = time.time()
    job_id = ('-').join(command_array).replace('/','-') # TODO get a unique job id
    job_id = re.sub(r'[^a-zA-Z0-9]', '-', job_id)
    task_id = re.sub(r'[^a-zA-Z0-9]', '-', task_id)
    std_out_log_name = f'{job_id}-{task_id}-{log_root_name}-stdout.log'
    std_err_log_name = f'{job_id}-{task_id}-{log_root_name}-stderr.log'
    with open(std_out_log_name,'w') as so:
        with open(std_err_log_name,'w') as se:
            with contextlib.redirect_stdout(so):
                with contextlib.redirect_stderr(se):
                    logging.info('-----INVOKING TASK-----------------------------------')
                    logging.info(f'Task ID: {task_id}')
                    logging.info(f'Command: {command_array}')    
                    result = subprocess.run(command_array, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=os.environ.copy())
                    output = result.stdout.decode('utf-8')
                    logging.info("Output:", output)
                    logging.info("Return code:", result.returncode)
    cos_conn.put(std_out_log_name,os.path.join(log_folder,std_out_log_name))
    cos_conn.put(std_err_log_name,os.path.join(log_folder,std_err_log_name))
    os.remove(std_out_log_name)
    os.remove(std_err_log_name)


# append processed to each line
def grid_process(source_file, target_file):
    with open(source_file, 'r') as src, open(target_file, 'w') as tgt:
        for line in src:
            processed_line = line.strip() + ' processed\n'
            tgt.write(processed_line)


"""
component_simple_grid_wrapper_example got wrapped by grid_wrapper, which wraps any CLAIMED component and implements the generic grid computing pattern https://romeokienzler.medium.com/the-generic-grid-computing-pattern-transforms-any-sequential-workflow-step-into-a-transient-grid-c7f3ca7459c8
This simple grid wrapper just scans a folder and for each file the grid_process function is called. Locking is achieved the following way:
Given source file1.ext is processed, simple_grid_wrapper creates files in the target_directory following the pattern file1.{STATUS}.ext where STATUS in:
LOCKED
PROCESSED
FAILED


CLAIMED component description: component-simple-grid-wrapper-example
"""

# pip install pandas

# component dependencies
# 

import os
import json
import random
import logging
import time
import glob
from pathlib import Path
import pandas as pd

# import component code
from component_simple_grid_wrapper_example import *


#folder containing input data in single files
sgw_source_folder = os.environ.get('sgw_source_folder')

#folder to store the output data in single files. Default: sgw_source_folder, in case sgw_source_folder==sgw_target_folder, files containing .LOCKED., .PROCESSED., .FAILED. are ignored
sgw_target_folder = os.environ.get('sgw_target_folder', sgw_source_folder)

# component interface


def get_next_batch():
    files = os.listdir(sgw_source_folder)
    if sgw_source_folder == sgw_target_folder:
        files = [
            f for f in files
            if not any(keyword in f for keyword in ["LOCKED", "PROCESSED", "FAILED"])
        ]

    # Filter files and check if corresponding target file exists
    filtered_files = []
    for file in files:
        file_name, file_ext = os.path.splitext(file)

        # Create target file names with LOCKED, PROCESSED, FAILED extensions
        target_file_locked = f"{file_name}.LOCKED{file_ext}"
        target_file_processed = f"{file_name}.PROCESSED{file_ext}"
        target_file_failed = f"{file_name}.FAILED{file_ext}"

        # Check if any of the target files exists
        if not any(
                os.path.exists(os.path.join(sgw_target_folder, target_file))
                for target_file in [target_file_locked, target_file_processed, target_file_failed]
        ):
            filtered_files.append(file)

    if filtered_files:
        return random.choice(filtered_files)
    else:
        return None


def process_wrapper(sub_process):
    sgw_target_folder_path = Path(sgw_target_folder)
    sgw_target_folder_path.mkdir(exist_ok=True, parents=True)

    while True:
        file_to_process = get_next_batch()
        logging.info(f"Processing batch: {file_to_process}")
        if file_to_process is None:
            break

        file_name = Path(file_to_process).stem
        file_ext = Path(file_to_process).suffix
        locked_file = sgw_target_folder+f"/{file_name}.LOCKED{file_ext}"
        locked_file_path = Path(locked_file)

        try:
            locked_file_path.touch()
            sub_process(sgw_source_folder +'/'+ file_to_process, locked_file)
            processed_file = sgw_target_folder+f"/{file_name}.PROCESSED{file_ext}"
            locked_file_path.rename(processed_file)

        except Exception as e:
            failed_file = sgw_target_folder+f"/{file_name}.FAILED{file_ext}"
            locked_file_path.rename(failed_file)

            with open(failed_file, 'w') as f:
                f.write(f"Exception occurred: {str(e)}\n")

            logging.error(f"Processing failed for {file_to_process}: {str(e)}")

    logging.info("Finished processing all batches.")


if __name__ == '__main__':
    process_wrapper(grid_process) 



import argparse
import os
import sys
import logging
import subprocess
import re
from c3.create_operator import create_cwl_component
from c3.pythonscript import Pythonscript
from c3.templates import component_setup_code_wo_logging, python_component_setup_code

def create_containerless_operator(
        file_path,
        version,
        skip_logging = False
    ):

    if version is None:
        version = 'latest'

    logging.debug(f'Called create_containerless_operator {version} with {file_path}')

    filename, file_extension = os.path.splitext(file_path)

    if file_extension != '.py':
        raise NotImplementedError('Containerless operators currenly only support python scripts')
    
    all_pip_packages_found = ''
    with open(file_path, 'r') as file:
        for line in file:
            if re.search('pip ', line):
                pip_packages = re.sub('[#, ,!]*pip[ ]*install[ ]*', '', line)
                logging.debug(f'PIP packages found: {pip_packages}')
                all_pip_packages_found += (f' {pip_packages}')
    logging.info(f'all PIP packages found: {all_pip_packages_found}')


    # prepend init code to script
    target_code = 'runnable.py'

    if os.path.exists(target_code):
        os.remove(target_code)

    with open(file_path, 'r') as f:
        script = f.read()
    if skip_logging:
        script = component_setup_code_wo_logging + script
    else:
        script = python_component_setup_code + script
    with open(target_code, 'w') as f:
        f.write(script)

    subprocess.run(';'.join(['rm -Rf claimedenv','python -m venv claimedenv', 
                                        'source ./claimedenv/bin/activate', 
                                        f'pip install {all_pip_packages_found.strip()}',
                                        'pip list',
                                        f'zip -r claimed-{filename}:{version}.zip {target_code} claimedenv',
                                        'rm -Rf claimedenv',
                                        f'rm {target_code}']), shell=True)
    script_data = Pythonscript(file_path)
    inputs = script_data.get_inputs()
    outputs = script_data.get_outputs()

    create_cwl_component(filename, "containerless", version, file_path, inputs, outputs)




def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('FILE_PATH', type=str,
                        help='Path to python script or notebook')
    parser.add_argument('ADDITIONAL_FILES', type=str, nargs='*', default=None,
                        help='Paths to additional files to include in the container image')
    parser.add_argument('-v', '--version', type=str, default=None,
                        help='Container image version. Auto-increases the version number if not provided (default 0.1)')
    parser.add_argument('-l', '--log_level', type=str, default='INFO')
    args = parser.parse_args()

    # Init logging
    root = logging.getLogger()
    root.setLevel(args.log_level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(args.log_level)
    root.addHandler(handler)

    create_containerless_operator(
        file_path=args.FILE_PATH,
        version=args.version,
    )

if __name__ == '__main__':
    main()
